--[[
-- Debris Class
-- Debris are the little pieces that fall when a Block, a Guardian or the Player are destroyed by an Explosion
-- These little pieces bounce around (using collision:getBounce()) until their internal timer (self.lived)
-- expires. Then they simply disappear.
--]]

local Debris = Entity:extend("Debris")

local minSize = .5
local maxSize = 2.5
local minVel = -50
local maxVel = (-1 * minVel)
local bounciness = 0.1

--something is wrong here, runs slooow, freezes game
function Debris:__init__(kwargs)
  kwargs.w,kwargs.h = 
    math.random(minSize, maxSize), math.random(minSize, maxSize)
  Entity.__init__(self,kwargs)
  self.elasticity = bounciness

  self.r, self.g, self.b = kwargs.r,kwargs.g,kwargs.b
  if kwargs.color then
      local col = getColor(kwargs.color)
      self.r, self.g, self.b = col[1],col[2],col[3]
  end
  self.lifeTime = 1 + 3 * math.random()
  self.lived = 0
  local minVel = kwargs.minVel or minVel
  local maxVel = kwargs.maxVel or maxVel
  self.vx = math.random(minVel, maxVel)
  self.vy = math.random(minVel, maxVel) --log("woo")
  self.isBackground = true
    if not self.passive then
        --hitchhiking
        self.map:storeBullet(self)
    end
end

function Debris:filter(other)
  if other.isDeubris or not other.isBackground then return "bounce" end
end

function Debris:moveColliding(dt)
  local world = self.world

  local future_l = self.x + self.vx * dt
  local future_t = self.y + self.vy * dt

  local next_l, next_t, cols, len = world:move(self, future_l, future_t, self.filter)
  local col
  for i=1, len do
    col = cols[i]
    self:resolveCollision(col)
  end

  self.x, self.y = next_l, next_t
end

function Debris:update(dt)
  self.lived = self.lived + dt

  if self.lived >= self.lifeTime  then
    self:destroy()
  else
    self:updateGravity(dt)
    self:moveColliding(dt)
  end
end

function Debris:draw()
  util.drawFilledRectangle(self.x, self.y, self.w, self.h, self.r, self.g, self.b)
end

return Debris
