local Map = class:extend("Map")

function Map:__init__(kwargs)
    local kwargs = kwargs or {}
    self.x = kwargs.x or 0
    self.y = kwargs.y or 0
    
    self.camera = Camera()
    
    self.h = kwargs.h or 10
    self.w = kwargs.w or self.h
    
    self.tH = kwargs.tH or 16
    self.tW = kwargs.tW or self.tH
    
    self.world = bump.newWorld()
    self.grid = {}
    
    self.diggers = {}
    self.entities = {}
    self.bullets = {}
    self.turrets = {}
    self.crates = {}
    self.toSet = {"x","y","life","vx","vy","source","w","h","angle"}
    
    self.camera.scale=2
    self.cameraMan = {x=0,y=0}
    self.vintage = true--cuts some costs and makes game run faster
end

--cuts all connections to entity
function Map:kill(entity)
    self.entities[entity.uuid] = nil
    --lume.remove(self.bullets,entity)
    self.bullets[entity.uuid] = nil
end

function Map:storeBullet(ent)
    if self.active then
        --table.insert(self.bullets,ent)
        self.entities[ent.uuid] = nil
        self.bullets[ent.uuid] = ent
    end
end

function Map:build()
    for x = 1,self.w do
        self.grid[x] = self.grid[x] or {}
        for y = 1,self.h do
            tile = {
                x = x,
                y = y,
                source = "dirt.png",
                isTile = true,
                isBackground = true,
                _x = x*self.tW,
                _y = y*self.tH,
                unUnit = null,
                unit = null,
                timeMade = love.timer.getTime()
            }
            if x==1 or x==self.w or y==1 or y==self.h then
                tile.isBackground = false
                tile.source = "metal.png"
                tile.timeMade = -1
            end
            
            self.world:add(tile,tile._x,tile._y,self.tW,self.tH)
            
            self.grid[x][y] = tile
        end
    end
    local t = self:getTile(self.w/2,self.h/2)
    self.camera:follow(t._x,t._y)
end

function Map:getTile(x,y)
    return self.grid[x] and self.grid[x][y]
end

function sort(self,b)
    return self.timeMade<b.timeMade
end

function Map:draw()
    self.camera:attach()
    local visibleThings, len = self.world:queryRect(self.camera:getVisible())
    table.sort(visibleThings,sort)

    for i=1, len do
        obj = visibleThings[i]
        if not obj.isTile then
            obj:draw()
        else
            self:drawTile(obj)
        end
    end
    if doDebug then
        drawBump(self.world)
    end
    self.camera:detach()
    self.camera:draw()
    if doDebug then
        drawBump(self.world,W()/4,10)
        love.graphics.print(inspect(love.graphics.getStats()),100,H()-450,0,2,2)
        love.graphics.print(string.format("FPS: %s",love.timer.getFPS()),100, H()-150,0,2,2)
        love.graphics.print(collectgarbage("count")/1024,100, H()-100,0,2,2)
        love.graphics.print(love.graphics.getStats().texturememory/(1024*1024),100,H()-50,0,2,2)
    end
    love.graphics.print("Bouncy Bullets: "..tlen(self.entities), W()/2,10)
end

function Map:update(dt)
    self.camera:update(dt)
    local visibleThings, len = self.world:queryRect(self.camera:getVisible())
    --table.sort(visibleThings)

    for i=1, len do
        obj = visibleThings[i]
        if not obj.isTile then
            obj:update(dt)
        end
    end
            
end

function Map:drawTile(tile)
    local w,h = resizeImage(getSource(tile.source),self.tW,self.tH)
    love.graphics.draw(getSource(tile.source),tile._x,tile._y,0,w,h)
end

function Map:mousereleased(x,y)
    local x,y = self.camera:toWorldCoords(x,y)
    local sx,sy = self:convertPixelToTile(x,y)
    if not self:getTile(sx,sy) then
        return
    end
    local bullet = bullets.BouncyBullet:new({
        x=x,
        y=y,
        map=self,
        parent = self
    })
end

function Map:convertPixelToTile(x,y)
    local xx,yy = floor(x/self.tW),
        floor(y/self.tH)
    if xx<1 then xx = xx+1 end
    if yy<1 then yy = yy+1 end
    return xx,yy
end

function Map:convertTileToPixel(x,y)
    return x*self.tW,
        y*self.tH
end

return Map